@@ -104,10 +104,10 @@ class SendTask(Task):
         self.target = target() if isinstance(target, Target) else target
         self.message = message
         self.deadline = deadline
-        self.retry = retry
         self.wait_for_ack = wait_for_ack
         self.service = SERVICE_NOTIFY if notification else SERVICE_RPC
         self.timer = None
+        self._retry = None if retry is None or retry < 0 else retry
         self._wakeup = threading.Event()
         self._error = None
 
@@ -122,18 +122,15 @@ class SendTask(Task):
         """Called immediately before the message is handed off to the i/o
         system.  This implies that the sender link is up.
         """
-        if not self.wait_for_ack:
-            # sender is not concerned with waiting for acknowledgment
-            # "best effort at-most-once delivery"
-            self._cleanup()
-            self._wakeup.set()
+        pass
 
     def _on_ack(self, state, info):
-        """Called by eventloop thread when the ack/nack is received from the
-        peer.
+        """If wait_for_ack is True, this is called by the eventloop thread when
+        the ack/nack is received from the peer.  If wait_for_ack is False this
+        is called by the eventloop right after the message is written to the
+        link.  In the last case state will always be set to ACCEPTED.
         """
         if state != pyngus.SenderLink.ACCEPTED:
-            # TODO(kgiusti): could retry if deadline not hit
             msg = ("{name} message send to {target} failed: remote"
                    " disposition: {disp}, info:"
                    "{info}".format(name=self.name,
@@ -179,15 +176,23 @@ class SendTask(Task):
             self.timer.cancel()
             self.timer = None
 
+    @property
+    def _can_retry(self):
+        # has the retry count expired?
+        if self._retry is not None:
+            self._retry -= 1
+            if self._retry < 0:
+                return False
+        return True
+
 
 class RPCCallTask(SendTask):
     """Performs an RPC Call.  Sends the request and waits for a response from
     the destination.
     """
-
-    def __init__(self, target, message, deadline, retry, wait_for_ack):
+    def __init__(self, target, message, deadline, retry):
         super(RPCCallTask, self).__init__("RPC Call", message, target,
-                                          deadline, retry, wait_for_ack)
+                                          deadline, retry, wait_for_ack=True)
         self._reply_link = None
         self._reply_msg = None
         self._msg_id = None
@@ -198,32 +203,30 @@ class RPCCallTask(SendTask):
 
     def _prepare(self, sender):
         # reserve a message id for mapping the received response
+        if self._msg_id:
+            # already set so this is a re-transmit. To be safe cancel the old
+            # msg_id and allocate a fresh one.
+            self._reply_link.cancel_response(self._msg_id)
         self._reply_link = sender._reply_link
         rl = self._reply_link
         self._msg_id = rl.prepare_for_response(self.message, self._on_reply)
 
     def _on_reply(self, message):
         # called if/when the reply message arrives
-        if self._wakeup.is_set():
-            LOG.debug("RPC Reply received after call completed")
-            return
         self._reply_msg = message
-        self._reply_link = None
+        self._msg_id = None  # to prevent _cleanup() from cancelling it
         self._cleanup()
         self._wakeup.set()
 
     def _on_ack(self, state, info):
-        if self._wakeup.is_set():
-            LOG.debug("RPC ACKed after call completed: %s %s", state, info)
-            return
         if state != pyngus.SenderLink.ACCEPTED:
             super(RPCCallTask, self)._on_ack(state, info)
         # must wait for reply if ACCEPTED
 
     def _cleanup(self):
-        if self._reply_link and self._msg_id:
+        if self._msg_id:
             self._reply_link.cancel_response(self._msg_id)
-            self._msg_id = None
+        self._reply_link = None
         super(RPCCallTask, self)._cleanup()
 
 
@@ -260,18 +263,23 @@ class Sender(pyngus.SenderEventHandler):
         self._address = None
         self._link = None
         self._scheduler = scheduler
-        self._delay = delay  # for re-connecting
+        self._delay = delay  # for re-connecting/re-transmitting
         # holds all pending SendTasks
         self._pending_sends = collections.deque()
         # holds all messages sent but not yet acked
         self._unacked = set()
         self._reply_link = None
         self._connection = None
+        self._resend_timer = None
 
     @property
     def pending_messages(self):
         return len(self._pending_sends)
 
+    @property
+    def unacked_messages(self):
+        return len(self._unacked)
+
     def attach(self, connection, reply_link, addresser):
         """Open the link. Called by the Controller when the AMQP connection
         becomes active.
@@ -290,10 +298,13 @@ class Sender(pyngus.SenderEventHandler):
         LOG.debug("Sender %s detached", self._address)
         self._connection = None
         self._reply_link = None
+        if self._resend_timer:
+            self._resend_timer.cancel()
+            self._resend_timer = None
         if self._link:
             self._link.close()
 
-    def reset(self):
+    def reset(self, reason="Link reset"):
         """Called by the controller on connection failover. Release all link
         resources, abort any in-flight messages, and check the retry limit on
         all pending send requests.
@@ -304,16 +315,16 @@ class Sender(pyngus.SenderEventHandler):
         if self._link:
             self._link.destroy()
             self._link = None
-        self._abort_unacked("Link reset")
-        self._check_retry_limit()
+        self._abort_unacked(reason)
+        self._check_retry_limit(reason)
 
-    def destroy(self):
+    def destroy(self, reason="Link destroyed"):
         """Destroy the sender and all pending messages.  Called on driver
         shutdown.
         """
         LOG.debug("Sender %s destroyed", self._address)
-        self.reset()
-        self._abort_pending("Link destroyed")
+        self.reset(reason)
+        self._abort_pending(reason)
 
     def send_message(self, send_task):
         """Send a message out the link.
@@ -354,30 +365,31 @@ class Sender(pyngus.SenderEventHandler):
         # sender_closed() will be called once the link completes closing
 
     def sender_closed(self, sender_link):
-        self._abort_unacked("Sender closed")
-        if self._connection:
-            # still attached, so attempt to restart the link
-            self._check_retry_limit()
-            self._scheduler.defer(self._reopen_link, self._delay)
+        self._handle_sender_closed()
 
     def sender_failed(self, sender_link, error):
         """Protocol error occurred."""
         LOG.warning(_LW("sender %(addr)s failed error=%(error)s"),
                     {'addr': self._address, 'error': error})
-        self.sender_closed(sender_link)
+        self._handle_sender_closed(str(error))
 
     # end Pyngus callbacks
 
-    def _check_retry_limit(self):
+    def _handle_sender_closed(self, reason="Sender closed"):
+        self._abort_unacked(reason)
+        if self._connection:
+            # still attached, so attempt to restart the link
+            self._check_retry_limit(reason)
+            self._scheduler.defer(self._reopen_link, self._delay)
+
+    def _check_retry_limit(self, reason):
         # Called on recoverable connection or link failure.  Remove any pending
         # sends that have exhausted their retry count:
         expired = set()
         for send_task in self._pending_sends:
-            if send_task.retry is not None:
-                send_task.retry -= 1
-                if send_task.retry <= 0:
-                    expired.add(send_task)
-                    send_task._on_error("Send retries exhausted")
+            if not send_task._can_retry:
+                expired.add(send_task)
+                send_task._on_error("Message send failed: %s" % reason)
         while expired:
             self._pending_sends.remove(expired.pop())
 
@@ -398,26 +410,75 @@ class Sender(pyngus.SenderEventHandler):
     def _can_send(self):
         return self._link and self._link.active
 
+    # acknowledge status
+    _TIMED_OUT = pyngus.SenderLink.TIMED_OUT
+    _ACCEPTED = pyngus.SenderLink.ACCEPTED
+    _RELEASED = pyngus.SenderLink.RELEASED
+    _MODIFIED = pyngus.SenderLink.MODIFIED
+
     def _send(self, send_task):
         send_task._prepare(self)
         send_task.message.address = self._address
+        if send_task.wait_for_ack:
+            self._unacked.add(send_task)
 
-        def pyngus_callback(link, handle, state, info):
-            # invoked when the message bus (n)acks this message
-            if state == pyngus.SenderLink.TIMED_OUT:
-                # ignore pyngus timeout - we maintain our own timer
-                return
-            self._unacked.discard(send_task)
-            send_task._on_ack(state, info)
+            def pyngus_callback(link, handle, state, info):
+                # invoked when the message bus (n)acks this message
+                if state == Sender._TIMED_OUT:
+                    # ignore pyngus timeout - we maintain our own timer
+                    # which will properly deal with this case
+                    return
+                self._unacked.discard(send_task)
+                if state == Sender._ACCEPTED:
+                    send_task._on_ack(Sender._ACCEPTED, info)
+                elif (state == Sender._RELEASED
+                      or (state == Sender._MODIFIED and
+                          # assuming delivery-failed means in-doubt:
+                          not info.get("delivery-failed") and
+                          not info.get("undeliverable-here"))):
+                    # These states indicate that the message was never
+                    # forwarded beyond the next hop so they can be
+                    # re-transmitted without risk of duplication
+                    self._resend(send_task)
+                else:
+                    # some error - let task figure it out...
+                    send_task._on_ack(state, info)
 
-        self._unacked.add(send_task)
-        self._link.send(send_task.message,
-                        delivery_callback=pyngus_callback,
-                        handle=self,
-                        deadline=send_task.deadline)
+            self._link.send(send_task.message,
+                            delivery_callback=pyngus_callback,
+                            handle=self,
+                            deadline=send_task.deadline)
+        else:  # do not wait for ack
+            self._link.send(send_task.message,
+                            delivery_callback=None,
+                            handle=self,
+                            deadline=send_task.deadline)
+            send_task._on_ack(pyngus.SenderLink.ACCEPTED, {})
+
+    def _resend(self, send_task):
+        # the message bus returned the message without forwarding it. Wait a
+        # bit for other outstanding sends to finish - most likely ending up
+        # here since they are all going to the same destination - then resend
+        # this message
+        if send_task._can_retry:
+            # note well: once there is something on the pending list no further
+            # messages will be sent (they will all queue up behind this one).
+            self._pending_sends.append(send_task)
+            if self._resend_timer is None:
+                sched = self._scheduler
+                # this will get the pending sends going again
+                self._resend_timer = sched.defer(self._resend_pending,
+                                                 self._delay)
+        else:
+            send_task._on_error("Send retries exhausted")
+
+    def _resend_pending(self):
+        # run from the _resend_timer, attempt to resend pending messages
+        self._resend_timer = None
+        self._send_pending()
 
     def _send_pending(self):
-        # send all pending messages
+        # flush all pending messages out
         if self._can_send:
             while self._pending_sends:
                 self._send(self._pending_sends.popleft())
@@ -469,7 +530,7 @@ class Replies(pyngus.ReceiverEventHandler):
             self._receiver.close()
 
     def destroy(self):
-        self._correlation = None
+        self._correlation.clear()
         if self._receiver:
             self._receiver.destroy()
             self._receiver = None
@@ -491,11 +552,10 @@ class Replies(pyngus.ReceiverEventHandler):
         """Abort waiting for the response message corresponding to msg_id.
         This can be used if the request fails and no reply is expected.
         """
-        if self._correlation:
-            try:
-                del self._correlation[msg_id]
-            except KeyError:
-                pass
+        try:
+            del self._correlation[msg_id]
+        except KeyError:
+            pass
 
     @property
     def active(self):
@@ -813,7 +873,7 @@ class Controller(pyngus.ConnectionEventHandler):
         self._closing = False
         # only schedule one outstanding reconnect attempt at a time
         self._reconnecting = False
-        self._delay = 1  # seconds between retries
+        self._delay = self.conn_retry_interval  # seconds between retries
         # prevent queuing up multiple requests to run _process_tasks()
         self._process_tasks_scheduled = False
         self._process_tasks_lock = threading.Lock()
@@ -843,7 +903,7 @@ class Controller(pyngus.ConnectionEventHandler):
             self.processor.wakeup(self._start_shutdown)
             LOG.debug("Waiting for eventloop to exit")
             self.processor.join(timeout)
-            self._hard_reset()
+            self._hard_reset("Shutting down")
             for sender in itervalues(self._all_senders):
                 sender.destroy()
             self._all_senders.clear()
@@ -861,8 +921,6 @@ class Controller(pyngus.ConnectionEventHandler):
         if send_task.deadline and send_task.deadline <= now():
             send_task._on_timeout()
             return
-        if send_task.retry is None or send_task.retry < 0:
-            send_task.retry = None
         key = keyify(send_task.target, send_task.service)
         sender = self._all_senders.get(key)
         if not sender:
@@ -870,7 +928,7 @@ class Controller(pyngus.ConnectionEventHandler):
                             self.link_retry_delay, send_task.service)
             self._all_senders[key] = sender
             if self.reply_link and self.reply_link.active:
-                sender.attach(self._socket_connection.connection,
+                sender.attach(self._socket_connection.pyngus_conn,
                               self.reply_link, self.addresser)
         self._active_senders.add(key)
         sender.send_message(send_task)
@@ -901,7 +959,7 @@ class Controller(pyngus.ConnectionEventHandler):
             self._servers[key] = servers
         servers[subscribe_task._subscriber_id] = server
         if self._active:
-            server.attach(self._socket_connection.connection,
+            server.attach(self._socket_connection.pyngus_conn,
                           self.addresser)
 
     # commands executed on the processor (eventloop) via 'wakeup()':
@@ -920,14 +978,9 @@ class Controller(pyngus.ConnectionEventHandler):
         if self.ssl_ca_file:
             conn_props["x-ssl-ca-file"] = self.ssl_ca_file
         if self.ssl_cert_file:
-            # assume this connection is for a server.  If client authentication
-            # support is developed, we'll need an explicit flag (server or
-            # client)
-            conn_props["x-ssl-server"] = True
             conn_props["x-ssl-identity"] = (self.ssl_cert_file,
                                             self.ssl_key_file,
                                             self.ssl_key_password)
-            conn_props["x-ssl-allow-cleartext"] = self.ssl_allow_insecure
         # SASL configuration:
         if self.sasl_mechanisms:
             conn_props["x-sasl-mechs"] = self.sasl_mechanisms
@@ -980,7 +1033,7 @@ class Controller(pyngus.ConnectionEventHandler):
             self._detach_senders()
             self._detach_servers()
             self.reply_link.detach()
-            self._socket_connection.connection.close()
+            self._socket_connection.pyngus_conn.close()
         else:
             # don't wait for a close from the remote, may never happen
             self.processor.shutdown()
@@ -996,7 +1049,7 @@ class Controller(pyngus.ConnectionEventHandler):
                  {'hostname': self.hosts.current.hostname,
                   'port': self.hosts.current.port})
         for sender in itervalues(self._all_senders):
-            sender.attach(self._socket_connection.connection,
+            sender.attach(self._socket_connection.pyngus_conn,
                           self.reply_link, self.addresser)
 
     def _reply_link_down(self):
@@ -1005,7 +1058,7 @@ class Controller(pyngus.ConnectionEventHandler):
         if not self._closing:
             self._detach_senders()
             self._detach_servers()
-            self._socket_connection.connection.close()
+            self._socket_connection.pyngus_conn.close()
             # once closed, _handle_connection_loss() will initiate reconnect
 
     # callback from eventloop on socket error
@@ -1013,7 +1066,7 @@ class Controller(pyngus.ConnectionEventHandler):
     def socket_error(self, error):
         """Called by eventloop when a socket error occurs."""
         LOG.error(_LE("Socket failure: %s"), error)
-        self._handle_connection_loss()
+        self._handle_connection_loss(str(error))
 
     # Pyngus connection event callbacks (and their helpers), all invoked from
     # the eventloop thread:
@@ -1022,11 +1075,11 @@ class Controller(pyngus.ConnectionEventHandler):
         """This is a Pyngus callback, invoked by Pyngus when a non-recoverable
         error occurs on the connection.
         """
-        if connection is not self._socket_connection.connection:
+        if connection is not self._socket_connection.pyngus_conn:
             # pyngus bug: ignore failure callback on destroyed connections
             return
         LOG.debug("AMQP Connection failure: %s", error)
-        self._handle_connection_loss()
+        self._handle_connection_loss(str(error))
 
     def connection_active(self, connection):
         """This is a Pyngus callback, invoked by Pyngus when the connection to
@@ -1042,13 +1095,13 @@ class Controller(pyngus.ConnectionEventHandler):
         self.addresser = self.addresser_factory(props)
         for servers in itervalues(self._servers):
             for server in itervalues(servers):
-                server.attach(self._socket_connection.connection,
+                server.attach(self._socket_connection.pyngus_conn,
                               self.addresser)
-        self.reply_link = Replies(self._socket_connection.connection,
+        self.reply_link = Replies(self._socket_connection.pyngus_conn,
                                   self._reply_link_ready,
                                   self._reply_link_down,
                                   self._reply_credit)
-        self._delay = 1
+        self._delay = self.conn_retry_interval   # reset
         # schedule periodic maintenance of sender links
         self._link_maint_timer = self.processor.defer(self._purge_sender_links,
                                                       self._link_maint_timeout)
@@ -1061,7 +1114,7 @@ class Controller(pyngus.ConnectionEventHandler):
         """
         LOG.debug("AMQP connection closed.")
         # if the driver isn't being shutdown, failover and reconnect
-        self._handle_connection_loss()
+        self._handle_connection_loss("AMQP connection closed.")
 
     def connection_remote_closed(self, connection, reason):
         """This is a Pyngus callback, invoked by Pyngus when the peer has
@@ -1075,7 +1128,7 @@ class Controller(pyngus.ConnectionEventHandler):
         self._detach_senders()
         self._detach_servers()
         self.reply_link.detach()
-        self._socket_connection.connection.close()
+        self._socket_connection.pyngus_conn.close()
 
     def sasl_done(self, connection, pn_sasl, outcome):
         """This is a Pyngus callback invoked when the SASL handshake
@@ -1089,13 +1142,14 @@ class Controller(pyngus.ConnectionEventHandler):
                   {'hostname': self.hosts.current.hostname,
                    'port': self.hosts.current.port,
                    'username': self.hosts.current.username})
-        # connection failure will be handled later
+        # pyngus will invoke connection_failed() eventually
 
-    def _handle_connection_loss(self):
+    def _handle_connection_loss(self, reason):
         """The connection to the messaging service has been lost.  Try to
         reestablish the connection/failover if not shutting down the driver.
         """
         self.addresser = None
+        self._socket_connection.close()
         if self._closing:
             # we're in the middle of shutting down the driver anyways,
             # just consider it done:
@@ -1107,43 +1161,45 @@ class Controller(pyngus.ConnectionEventHandler):
                 self._reconnecting = True
                 LOG.info(_LI("delaying reconnect attempt for %d seconds"),
                          self._delay)
-                self.processor.defer(self._do_reconnect, self._delay)
-                self._delay = min(self._delay * 2, 60)
+                self.processor.defer(lambda: self._do_reconnect(reason),
+                                     self._delay)
+                self._delay = min(self._delay * self.conn_retry_backoff,
+                                  self.conn_retry_interval_max)
             if self._link_maint_timer:
                 self._link_maint_timer.cancel()
                 self._link_maint_timer = None
 
-    def _do_reconnect(self):
+    def _do_reconnect(self, reason):
         """Invoked on connection/socket failure, failover and re-connect to the
         messaging service.
         """
         self._reconnecting = False
         if not self._closing:
-            self._hard_reset()
+            self._hard_reset(reason)
             host = self.hosts.next()
             LOG.info(_LI("Reconnecting to: %(hostname)s:%(port)s"),
                      {'hostname': host.hostname, 'port': host.port})
             self._socket_connection.connect(host)
 
-    def _hard_reset(self):
+    def _hard_reset(self, reason):
         """Reset the controller to its pre-connection state"""
         # note well: since this method destroys the connection, it cannot be
         # invoked directly from a pyngus callback.  Use processor.defer() to
         # run this method on the main loop instead.
         for sender in self._purged_senders:
-            sender.destroy()
+            sender.destroy(reason)
         del self._purged_senders[:]
         self._active_senders.clear()
         unused = []
         for key, sender in iteritems(self._all_senders):
-            # clean up any unused sender links
+            # clean up any sender links that no longer have messages to send
             if sender.pending_messages == 0:
                 unused.append(key)
             else:
-                sender.reset()
+                sender.reset(reason)
                 self._active_senders.add(key)
         for key in unused:
-            self._all_senders[key].destroy()
+            self._all_senders[key].destroy(reason)
             del self._all_senders[key]
         for servers in itervalues(self._servers):
             for server in itervalues(servers):
@@ -1170,14 +1226,14 @@ class Controller(pyngus.ConnectionEventHandler):
         if not self._closing:
             # destroy links that have already been closed
             for sender in self._purged_senders:
-                sender.destroy()
+                sender.destroy("Idle link purged")
             del self._purged_senders[:]
 
             # determine next set to purge
             purge = set(self._all_senders.keys()) - self._active_senders
             for key in purge:
                 sender = self._all_senders[key]
-                if sender.pending_messages == 0:
+                if not sender.pending_messages and not sender.unacked_messages:
                     sender.detach()
                     self._purged_senders.append(self._all_senders.pop(key))
             self._active_senders.clear()
@@ -1189,4 +1245,4 @@ class Controller(pyngus.ConnectionEventHandler):
     def _active(self):
         # Is the connection up
         return (self._socket_connection
-                and self._socket_connection.connection.active)
+                and self._socket_connection.pyngus_conn.active)
