@@ -15,10 +15,12 @@
 import logging
 import os
 import select
+import shlex
 import shutil
 import socket
 import subprocess
 import sys
+import tempfile
 import threading
 import time
 import uuid
@@ -49,6 +51,8 @@ if pyngus:
 _proton = importutils.try_import("proton")
 CYRUS_ENABLED = (pyngus and pyngus.VERSION >= (2, 0, 0) and _proton
                  and getattr(_proton.SASL, "extended", lambda: False)())
+# same with SSL
+SSL_ENABLED = (_proton and getattr(_proton.SSL, "present", lambda: False)())
 
 LOG = logging.getLogger(__name__)
 
@@ -288,7 +292,7 @@ class TestAmqpSend(_AmqpBrokerTestCaseAuto):
         driver.cleanup()
 
     def test_send_timeout(self):
-        """Verify send timeout."""
+        """Verify send timeout - no reply sent."""
         driver = amqp_driver.ProtonDriver(self.conf, self._broker_url)
         target = oslo_messaging.Target(topic="test-topic")
         listener = _ListenerThread(
@@ -310,17 +314,19 @@ class TestAmqpSend(_AmqpBrokerTestCaseAuto):
         driver = amqp_driver.ProtonDriver(self.conf, self._broker_url)
         target = oslo_messaging.Target(topic="no listener")
 
-        # the broker will send a nack:
+        # the broker will send a nack (released) since there is no active
+        # listener for the target:
         self.assertRaises(oslo_messaging.MessageDeliveryFailure,
                           driver.send, target,
                           {"context": "whatever"},
                           {"method": "drop"},
                           wait_for_reply=True,
+                          retry=0,
                           timeout=1.0)
         driver.cleanup()
 
     def test_send_not_acked(self):
-        """Verify exception thrown if send Nacked."""
+        """Verify exception thrown ack dropped."""
         self.config(pre_settled=[],
                     group="oslo_messaging_amqp")
         driver = amqp_driver.ProtonDriver(self.conf, self._broker_url)
@@ -333,7 +339,8 @@ class TestAmqpSend(_AmqpBrokerTestCaseAuto):
                           driver.send, target,
                           {"context": "whatever"},
                           {"method": "drop"},
-                          wait_for_reply=False)
+                          retry=0,
+                          wait_for_reply=True)
         driver.cleanup()
 
     def test_no_ack_cast(self):
@@ -393,7 +400,7 @@ class TestAmqpSend(_AmqpBrokerTestCaseAuto):
         driver.cleanup()
 
     def test_call_failed_reply(self):
-        """Send back an exception"""
+        """Send back an exception generated at the listener"""
         class _FailedResponder(_ListenerThread):
             def __init__(self, listener):
                 super(_FailedResponder, self).__init__(listener, 1)
@@ -434,7 +441,7 @@ class TestAmqpSend(_AmqpBrokerTestCaseAuto):
                 self.started.set()
                 while not self._done:
                     for in_msg in self.listener.poll(timeout=0.5):
-                        # reply will never be acked:
+                        # reply will never be acked (simulate drop):
                         in_msg._reply_to = "!no-ack!"
                         in_msg.reply(reply={'correlation-id':
                                             in_msg.message.get("id")})
@@ -458,6 +465,7 @@ class TestAmqpSend(_AmqpBrokerTestCaseAuto):
 
     def test_listener_requeue(self):
         "Emulate Server requeue on listener incoming messages"
+        self.config(pre_settled=[], group="oslo_messaging_amqp")
         driver = amqp_driver.ProtonDriver(self.conf, self._broker_url)
         driver.require_features(requeue=True)
         target = oslo_messaging.Target(topic="test-topic")
@@ -472,10 +480,6 @@ class TestAmqpSend(_AmqpBrokerTestCaseAuto):
         listener.join(timeout=30)
         self.assertFalse(listener.isAlive())
 
-        for x in listener.get_messages():
-            x.requeue()
-            self.assertEqual(x.message, {"msg": "value"})
-
         predicate = lambda: (self._broker.sender_link_requeue_count == 1)
         _wait_until(predicate, 30)
         self.assertTrue(predicate())
@@ -575,7 +579,7 @@ class TestAmqpNotification(_AmqpBrokerTestCaseAuto):
                 try:
                     driver.send_notification(oslo_messaging.Target(topic=t),
                                              "context", {'target': t},
-                                             version)
+                                             version, retry=0)
                 except oslo_messaging.MessageDeliveryFailure:
                     excepted_targets.append(t)
 
@@ -592,15 +596,18 @@ class TestAmqpNotification(_AmqpBrokerTestCaseAuto):
         driver.cleanup()
 
     def test_released_notification(self):
+        """Broker sends a Nack (released)"""
         driver = amqp_driver.ProtonDriver(self.conf, self._broker_url)
         self.assertRaises(oslo_messaging.MessageDeliveryFailure,
                           driver.send_notification,
                           oslo_messaging.Target(topic="bad address"),
                           "context", {'target': "bad address"},
-                          2.0)
+                          2.0,
+                          retry=0)
         driver.cleanup()
 
     def test_notification_not_acked(self):
+        """Simulate drop of ack from broker"""
         driver = amqp_driver.ProtonDriver(self.conf, self._broker_url)
         # set this directly so we can use a value < minimum allowed
         driver._default_notify_timeout = 2
@@ -608,7 +615,7 @@ class TestAmqpNotification(_AmqpBrokerTestCaseAuto):
                           driver.send_notification,
                           oslo_messaging.Target(topic="!no-ack!"),
                           "context", {'target': "!no-ack!"},
-                          2.0)
+                          2.0, retry=0)
         driver.cleanup()
 
     def test_no_ack_notification(self):
@@ -674,12 +681,12 @@ class TestAuthentication(test_utils.BaseTestCase):
         target = oslo_messaging.Target(topic="test-topic")
         _ListenerThread(
             driver.listen(target, None, None)._poll_style_listener, 1)
-        self.assertRaises(oslo_messaging.MessagingTimeout,
+        self.assertRaises(oslo_messaging.MessageDeliveryFailure,
                           driver.send,
                           target, {"context": True},
                           {"method": "echo"},
                           wait_for_reply=True,
-                          timeout=2.0)
+                          retry=2)
         driver.cleanup()
 
 
@@ -771,7 +778,6 @@ mech_list: ${mechs}
         """Verify that a bad password given in TransportHost is
         rejected by the broker.
         """
-
         addr = "amqp://joe:badpass@%s:%d" % (self._broker.host,
                                              self._broker.port)
         url = oslo_messaging.TransportURL.parse(self.conf, addr)
@@ -779,12 +785,15 @@ mech_list: ${mechs}
         target = oslo_messaging.Target(topic="test-topic")
         _ListenerThread(
             driver.listen(target, None, None)._poll_style_listener, 1)
-        self.assertRaises(oslo_messaging.MessagingTimeout,
-                          driver.send,
-                          target, {"context": True},
-                          {"method": "echo"},
-                          wait_for_reply=True,
-                          timeout=2.0)
+        try:
+            driver.send(target, {"context": True}, {"method": "echo"},
+                        wait_for_reply=True, retry=2)
+        except oslo_messaging.MessageDeliveryFailure as e:
+            # verify the exception indicates the failure was an authentication
+            # error
+            self.assertTrue('amqp:unauthorized-access' in str(e))
+        else:
+            self.assertIsNone("Expected authentication failure")
         driver.cleanup()
 
     def test_authentication_bad_mechs(self):
@@ -800,12 +809,12 @@ mech_list: ${mechs}
         target = oslo_messaging.Target(topic="test-topic")
         _ListenerThread(
             driver.listen(target, None, None)._poll_style_listener, 1)
-        self.assertRaises(oslo_messaging.MessagingTimeout,
+        self.assertRaises(oslo_messaging.MessageDeliveryFailure,
                           driver.send,
                           target, {"context": True},
                           {"method": "echo"},
                           wait_for_reply=True,
-                          timeout=2.0)
+                          retry=0)
         driver.cleanup()
 
     def test_authentication_default_username(self):
@@ -1386,6 +1395,206 @@ class TestAddressing(test_utils.BaseTestCase):
                               LegacyAddresser)
 
 
+@testtools.skipUnless(pyngus, "proton modules not present")
+class TestMessageRetransmit(_AmqpBrokerTestCase):
+    # test message is retransmitted if safe to do so
+    def _test_retransmit(self, nack_method):
+        self._nack_count = 2
+
+        def _on_message(message, handle, link):
+            if self._nack_count:
+                self._nack_count -= 1
+                nack_method(link, handle)
+            else:
+                self._broker.forward_message(message, handle, link)
+
+        self._broker.on_message = _on_message
+        self._broker.start()
+        self.config(link_retry_delay=1, pre_settled=[],
+                    group="oslo_messaging_amqp")
+        driver = amqp_driver.ProtonDriver(self.conf, self._broker_url)
+        target = oslo_messaging.Target(topic="test-topic")
+        listener = _ListenerThread(driver.listen(target,
+                                                 None,
+                                                 None)._poll_style_listener,
+                                   1)
+        rc = driver.send(target, {"context": "whatever"},
+                         {"method": "echo", "id": "blah"},
+                         wait_for_reply=True,
+                         retry=2)  # initial send + up to 2 resends
+        self.assertIsNotNone(rc)
+        self.assertEqual(0, self._nack_count)
+        self.assertEqual(rc.get('correlation-id'), 'blah')
+        listener.join(timeout=30)
+        self.assertFalse(listener.isAlive())
+        driver.cleanup()
+
+    def test_released(self):
+        # should retry and succeed
+        self._test_retransmit(lambda l, h: l.message_released(h))
+
+    def test_modified(self):
+        # should retry and succeed
+        self._test_retransmit(lambda l, h: l.message_modified(h,
+                                                              False,
+                                                              False,
+                                                              {}))
+
+    def test_modified_failed(self):
+        # since delivery_failed is set to True, should fail
+        self.assertRaises(oslo_messaging.MessageDeliveryFailure,
+                          self._test_retransmit,
+                          lambda l, h: l.message_modified(h, True, False, {}))
+
+    def test_rejected(self):
+        # rejected - should fail
+        self.assertRaises(oslo_messaging.MessageDeliveryFailure,
+                          self._test_retransmit,
+                          lambda l, h: l.message_rejected(h, {}))
+
+
+@testtools.skipUnless(SSL_ENABLED, "OpenSSL not supported")
+class TestSSL(test_utils.BaseTestCase):
+    """Test the driver's OpenSSL integration"""
+
+    def setUp(self):
+        super(TestSSL, self).setUp()
+        # Create the CA, server, and client SSL certificates:
+        self._tmpdir = tempfile.mkdtemp(prefix='amqp1')
+        files = ['ca_key', 'ca_cert', 's_key', 's_req', 's_cert', 'c_key',
+                 'c_req', 'c_cert', 'bad_cert', 'bad_req', 'bad_key']
+        conf = dict(zip(files, [os.path.join(self._tmpdir, "%s.pem" % f)
+                                for f in files]))
+        conf['pw'] = 'password'
+        conf['s_name'] = '127.0.0.1'
+        conf['c_name'] = 'client.com'
+        self._ssl_config = conf
+        ssl_setup = [
+            # create self-signed CA key and certificate:
+            Template('openssl genrsa -out ${ca_key} 2048').substitute(conf),
+            Template('openssl req -x509 -key ${ca_key} -subj'
+                     ' "/CN=Trusted.CA.com" -out'
+                     ' ${ca_cert}').substitute(conf),
+            # create Server key and certificate:
+            Template('openssl genrsa -out ${s_key} 2048').substitute(conf),
+            Template('openssl req -new -key ${s_key} -subj /CN=${s_name}'
+                     ' -passin pass:${pw} -out ${s_req}').substitute(conf),
+            Template('openssl x509 -req -in ${s_req} -CA ${ca_cert}'
+                     ' -CAkey ${ca_key} -CAcreateserial -out'
+                     ' ${s_cert}').substitute(conf),
+            # create a "bad" Server cert for testing CN validation:
+            Template('openssl genrsa -out ${bad_key} 2048').substitute(conf),
+            Template('openssl req -new -key ${bad_key} -subj /CN=Invalid'
+                     ' -passin pass:${pw} -out ${bad_req}').substitute(conf),
+            Template('openssl x509 -req -in ${bad_req} -CA ${ca_cert}'
+                     ' -CAkey ${ca_key} -CAcreateserial -out'
+                     ' ${bad_cert}').substitute(conf),
+            # create Client key and certificate for client authentication:
+            Template('openssl genrsa -out ${c_key} 2048').substitute(conf),
+            Template('openssl req -new -key ${c_key} -subj /CN=${c_name}'
+                     ' -passin pass:${pw} -out'
+                     ' ${c_req}').substitute(conf),
+            Template('openssl x509 -req -in ${c_req} -CA ${ca_cert}'
+                     ' -CAkey ${ca_key} -CAcreateserial -out'
+                     ' ${c_cert}').substitute(conf)
+        ]
+        for cmd in ssl_setup:
+            try:
+                subprocess.check_call(args=shlex.split(cmd))
+            except Exception:
+                shutil.rmtree(self._tmpdir, ignore_errors=True)
+                self._tmpdir = None
+                self.skipTest("OpenSSL tools not installed - skipping")
+
+    def test_server_ok(self):
+        # test client authenticates server
+        self._broker = FakeBroker(self.conf.oslo_messaging_amqp,
+                                  sock_addr=self._ssl_config['s_name'],
+                                  ssl_config=self._ssl_config)
+        url = oslo_messaging.TransportURL.parse(self.conf, "amqp://%s:%d" %
+                                                (self._broker.host,
+                                                 self._broker.port))
+        self._broker.start()
+
+        self.config(ssl_ca_file=self._ssl_config['ca_cert'],
+                    group='oslo_messaging_amqp')
+        driver = amqp_driver.ProtonDriver(self.conf, url)
+        target = oslo_messaging.Target(topic="test-topic")
+        listener = _ListenerThread(
+            driver.listen(target, None, None)._poll_style_listener, 1)
+
+        driver.send(target,
+                    {"context": "whatever"},
+                    {"method": "echo", "a": "b"},
+                    wait_for_reply=True,
+                    timeout=30)
+        listener.join(timeout=30)
+        self.assertFalse(listener.isAlive())
+        driver.cleanup()
+
+    def test_bad_server_fail(self):
+        # test client does not connect to invalid server
+        self._ssl_config['s_cert'] = self._ssl_config['bad_cert']
+        self._ssl_config['s_key'] = self._ssl_config['bad_key']
+        self._broker = FakeBroker(self.conf.oslo_messaging_amqp,
+                                  sock_addr=self._ssl_config['s_name'],
+                                  ssl_config=self._ssl_config)
+        url = oslo_messaging.TransportURL.parse(self.conf, "amqp://%s:%d" %
+                                                (self._broker.host,
+                                                 self._broker.port))
+        self._broker.start()
+
+        self.config(ssl_ca_file=self._ssl_config['ca_cert'],
+                    group='oslo_messaging_amqp')
+        driver = amqp_driver.ProtonDriver(self.conf, url)
+        target = oslo_messaging.Target(topic="test-topic")
+        self.assertRaises(oslo_messaging.MessageDeliveryFailure,
+                          driver.send, target,
+                          {"context": "whatever"},
+                          {"method": "echo", "a": "b"},
+                          wait_for_reply=False,
+                          retry=1)
+        driver.cleanup()
+
+    def test_client_auth_ok(self):
+        # test server authenticates client
+        self._ssl_config['authenticate_client'] = True
+        self._broker = FakeBroker(self.conf.oslo_messaging_amqp,
+                                  sock_addr=self._ssl_config['s_name'],
+                                  ssl_config=self._ssl_config)
+        url = oslo_messaging.TransportURL.parse(self.conf, "amqp://%s:%d" %
+                                                (self._broker.host,
+                                                 self._broker.port))
+        self._broker.start()
+
+        self.config(ssl_ca_file=self._ssl_config['ca_cert'],
+                    ssl_cert_file=self._ssl_config['c_cert'],
+                    ssl_key_file=self._ssl_config['c_key'],
+                    ssl_key_password=self._ssl_config['pw'],
+                    group='oslo_messaging_amqp')
+        driver = amqp_driver.ProtonDriver(self.conf, url)
+        target = oslo_messaging.Target(topic="test-topic")
+        listener = _ListenerThread(
+            driver.listen(target, None, None)._poll_style_listener, 1)
+
+        driver.send(target,
+                    {"context": "whatever"},
+                    {"method": "echo", "a": "b"},
+                    wait_for_reply=True,
+                    timeout=30)
+        listener.join(timeout=30)
+        self.assertFalse(listener.isAlive())
+        driver.cleanup()
+
+    def tearDown(self):
+        if self._broker:
+            self._broker.stop()
+            self._broker = None
+        if self._tmpdir:
+            shutil.rmtree(self._tmpdir, ignore_errors=True)
+        super(TestSSL, self).tearDown()
+
+
 class FakeBroker(threading.Thread):
     """A test AMQP message 'broker'."""
 
@@ -1403,6 +1612,7 @@ class FakeBroker(threading.Thread):
                 self.sasl_mechanisms = sasl_mechanisms
                 self.user_credentials = user_credentials
                 properties = {'x-server': True}
+                # setup SASL:
                 if self.sasl_mechanisms:
                     properties['x-sasl-mechs'] = self.sasl_mechanisms
                     if "ANONYMOUS" not in self.sasl_mechanisms:
@@ -1411,6 +1621,19 @@ class FakeBroker(threading.Thread):
                     properties['x-sasl-config-dir'] = sasl_config_dir
                 if sasl_config_name:
                     properties['x-sasl-config-name'] = sasl_config_name
+                # setup SSL
+                if self.server._ssl_config:
+                    ssl = self.server._ssl_config
+                    properties['x-ssl-server'] = True
+                    properties['x-ssl-identity'] = (ssl['s_cert'],
+                                                    ssl['s_key'],
+                                                    ssl['pw'])
+                    # check for client authentication
+                    if ssl.get('authenticate_client'):
+                        properties['x-ssl-ca-file'] = ssl['ca_cert']
+                        properties['x-ssl-verify-mode'] = 'verify-peer'
+                        properties['x-ssl-peer-name'] = ssl['c_name']
+                # misc connection properties
                 if product:
                     properties['properties'] = {'product': product}
 
@@ -1607,7 +1830,7 @@ class FakeBroker(threading.Thread):
 
             def message_received(self, receiver_link, message, handle):
                 """Forward this message out the proper sending link."""
-                self.server.forward_message(message, handle, receiver_link)
+                self.server.on_message(message, handle, receiver_link)
                 if self.link.capacity < 1:
                     self.server.on_credit_exhausted(self.link)
 
@@ -1618,17 +1841,18 @@ class FakeBroker(threading.Thread):
                  sasl_mechanisms="ANONYMOUS",
                  user_credentials=None,
                  sasl_config_dir=None,
-                 sasl_config_name=None):
+                 sasl_config_name=None,
+                 ssl_config=None):
         """Create a fake broker listening on sock_addr:sock_port."""
         if not pyngus:
             raise AssertionError("pyngus module not present")
         threading.Thread.__init__(self)
-        self._config = cfg
         self._product = product
         self._sasl_mechanisms = sasl_mechanisms
         self._sasl_config_dir = sasl_config_dir
         self._sasl_config_name = sasl_config_name
         self._user_credentials = user_credentials
+        self._ssl_config = ssl_config
         self._wakeup_pipe = os.pipe()
         self._my_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         self._my_socket.bind((sock_addr, sock_port))
@@ -1672,6 +1896,7 @@ class FakeBroker(threading.Thread):
         self.on_sender_active = lambda link: None
         self.on_receiver_active = lambda link: link.add_capacity(10)
         self.on_credit_exhausted = lambda link: link.add_capacity(10)
+        self.on_message = lambda m, h, l: self.forward_message(m, h, l)
 
     def start(self):
         """Start the server."""
